import numpy as np
from copy import deepcopy
import curves as c
#from curvepair import curvePair
import itertools as it
import sys
from numba import autojit,jit
'''
import logging

import numba.codegen.debug
#def disableNumbaLogging():

llvmlogger = logging.getLogger('numba.codegen.debug')
llvmlogger.setLevel(logging.INFO)
'''


def shift(path):
    '''
    init

    '''
    temp = path.index(min(path))
    return path[temp:] + path[:temp]

def invert(path):
    '''
    init

    '''
    return shift(path[::-1])

@jit
def dfs_wrapper(graph_obj, nodes):
    graph_copy = deepcopy(graph_obj.nodes)
    for start_node in nodes:
        for adj_node in graph_copy[start_node]:
            graph_obj.loopDFS(start_node,adj_node,graph_copy,[start_node],graph_obj.loops, graph_obj.nodes_to_faces)
rep_num = 1


class Graph:

    def addNode(self,node):
        self.nodes[node] = []


    def __init__(self, curve_pair, param_edges):

        self.nodes = {}
        self.counter = 0
        self.loops = []
        self.gammas = []
        self.nodes_to_faces = {}
        edges, __ =  curve_pair.edges
        fourgons = [i[1] for i in edges if i[0] == 4  ]
        non_fourgons = [i[1] for i in edges if i[0] != 4  ]
        keys = self.nodes_to_faces.keys()
        for i,face in enumerate(non_fourgons):
            for node in face:
                if not node in keys:
                    self.nodes_to_faces[node] = [i]
                else:
                    self.nodes_to_faces[node].append(i)
                keys = self.nodes_to_faces.keys()
        for face in fourgons:
            for node in face:
                if not node in keys:
                    self.nodes_to_faces[node] = [None]
                else:
                    self.nodes_to_faces[node].append(None)

                keys = self.nodes_to_faces.keys()
        nodes = range(curve_pair.n)

        for i in nodes:
            self.addNode(i)
        self.rep_num = 1
        self.findAllEdges(fourgons, non_fourgons, nodes, self.rep_num)
        graph_copy = deepcopy(self.nodes)
        '''

        import multiprocessing as mp
        procs = []
        for start_node in nodes:
            for adj_node in graph_copy[start_node]:
                proc = mp.Process(target=self.loopDFS, args=(start_node,adj_node,graph_copy,[start_node], self.loops, self.nodes_to_faces))
                proc.start()
                procs.append(proc)
            for proc in procs:
                proc.join()
            print self.loops
            procs = []
        '''
        '''

        for start_node in nodes:
            for adj_node in graph_copy[start_node]:
                self.loopDFS(start_node,adj_node,graph_copy,[start_node],self.loops, self.nodes_to_faces)
        '''
        from johnson import Johnson
        johnny = Johnson(graph_copy)
        johnny.find_all_circuits()
        self.loops = johnny.circuits
        #print self.loops
        #stderr.write(str(graph_copy)+'\n')
        #stderr.write(str(self.loops))
        self.loops = [list(j) for j in set([tuple(i) for i in self.loops])]
        __, edges = param_edges
        for path in list(self.loops):
            removed = False
            if len(path) < 3:
                if not removed:
                    self.loops.remove(path)
                    removed = True

            elif invert(path) in self.loops:
                if not removed:
                    self.loops.remove(path)
                    removed = True



            # Trial: remove all duplicates
            else:
                temp_path = list(path)
                temp_path = shift(temp_path)
                for face in non_fourgons:
                    for triple in [temp_path[i:i+3] \
                            for i in range(len(temp_path)-2)]:
                        if set(triple) <= set(face):
                            if not removed:
                                self.loops.remove(path)
                                removed = True

                    temp_path = invert(temp_path)

                    for triple in [temp_path[i:i+3] \
                            for i in range(len(temp_path)-2)]:
                        if set(triple) <= set(face):
                            if not removed:
                                self.loops.remove(path)
                                removed = True

                    for i in range(len(path)):
                        temp_path = temp_path[1:] + temp_path[:1]

                        for  triple in [temp_path[i:i+3] \
                                for i in range(len(temp_path)-2)]:
                            if set(triple) <= set(face) and path in self.loops:
                                if not removed:
                                    self.loops.remove(path)
                                    removed = True
        from curvepair import CurvePair
        for loop in list(self.loops):
            #beta = list([list(curvePair.top),list(curvePair.bottom)])
            path = list(loop)
            path2 = list(path)
            path_matrix = c.buildMatrices(deepcopy(edges), [path])
            ladder = [list(path_matrix[0][0,:,1]),list(path_matrix[0][0,:,3])]

            gamma = CurvePair(ladder[0],ladder[1],0, 0)
            if gamma.genus <= curve_pair.genus:
                self.gammas.append(loop)

    def getValue(self,pos_to_insert, ladder, path):
        return	int(path[int(ladder[int(pos_to_insert)])])+1

    def findAllEdges(self, fourgons, non_fourgons, alphaEdgeNodes, rep_num):
        '''
        Determines all edges between boundary componenets and adds adjacencies between
        appropriate edges.

        '''

        #find all direct connections between non-fourgon regions
        regions = fourgons + non_fourgons
        for alphaEdge in alphaEdgeNodes:
            for region in regions:
                if alphaEdge in region:
                    region.remove(alphaEdge)
                    for other_edge in region:
                        self.addAdjacency(alphaEdge, int(other_edge), rep_num)
                    region.add(alphaEdge)


    def addAdjacency(self,node, adjacentNode, rep_num):
        '''
        Adds adjacencies to a graph represented by an adjacency list.
        This is useful when we would like to replicate adjacencies
        between nodes.

        :param self:
        :type self: Graph
        :param node: node to add adjacency
        :param type: int
        :param adjacentNode: the adjacent node
        :type adjacentNode: int
        :param rep_num: set to 1
        :type rep_num: int

        '''
        for i in range(rep_num):
            adjacencyList = self.nodes[node]
            if Graph.count(adjacentNode, adjacencyList) < rep_num:
                adjacencyList.append(adjacentNode)

            adjacencyList = self.nodes[adjacentNode]
            if Graph.count(node, adjacencyList) < rep_num:
                adjacencyList.append(node)
    @staticmethod
    def count(adj_node, adj_list):
        '''
        Determines the number of adjacencies between two nodes in a graph, given the
        adjacency list of one of node.

        :param self:
        :type self: Graph
        :param adj_node: Adjacent node
        :type adj_node: int
        :param adj_list: Adjacency list of the node in question
        :type adj_list: list<int>
        :returns: number of edges from the adjacent node to the original node


        '''
        count = 0
        for i in adj_list:
            if i == adj_node:
                count += 1
        return count

    def loopDFS(self, current_node,start_node,graph,current_path,all_loops,nodes_to_faces):
        '''
        Recursively finds all closed cycles in a given graph that begin and end at start_node.
        As one would guess, it employs a standard depth-first search algorithm on the graph,
        appending current_path to all_loops when it returns to start_node.

        In the overall distance computation, this function is computationally dominating with
        exponential complexity, so take care with its use.

        :param self:
        :type self: Graph
        :param current_node: the current alpha edge in the recursion
        :type current_node: int
        :param start_node: the source node of the current recursive search
        :type start_node: int
        :param graph: graph of the overall graph mid-recursion
        :type graph: dict<int,list<int> >
        :param current_path: list of nodes in the current path
        :type current_path: list<int>
        :param all_loops: list of all current paths
        :type all_loops: list< list<int> >
        :returns: set of all closeds cycles in the graph starting and ending at start_node

        '''
        if len(current_path) >= 3:
            path_head_3 = current_path[-3:]
            #path_head_2 = current_path[-2:]
            #previous_three_faces = []
            #for edge in path_head_3:
            #	previous_three_faces.append(set(self.nodes_to_faces[edge]))

            previous_three_faces =  [set(nodes_to_faces[edge]) for edge in path_head_3]
            #previous_two_faces =  [set(self.nodes_to_faces[edge]) for edge in path_head_2]
            #print 'ptf:',previous_three_faces[0],previous_three_faces[1],previous_three_faces[2]
            #intersection_all = previous_three_faces[0]
            #intersection_all = intersection_all.intersection(previous_three_faces[1])
            #intersection_all = intersection_all.intersection(previous_three_faces[2])
            intersection_all = set.intersection(*previous_three_faces) #old non numba
            if len(intersection_all) == 2:
                return

        if current_node == start_node:
            all_loops.append(shift(list(current_path)))
            return

        else:
            for adjacent_node in set(graph[current_node]):
                if Graph.count(adjacent_node, current_path) < rep_num:
                    current_path.append(adjacent_node)
                    graph[current_node].remove(adjacent_node)
                    graph[adjacent_node].remove(current_node)
                    self.loopDFS(adjacent_node, start_node,deepcopy(graph), current_path, all_loops, nodes_to_faces)
                    graph[current_node].append(adjacent_node)
                    graph[adjacent_node].append(current_node)
                    current_path.pop()

'''
for i in range(len(path)):
    index = path[i]
    beta[0].insert(index+1,None)
    beta[1].insert(index+1,None)
    beta[0].append('____')
    beta[1].append('____')
    for j,tup in enumerate(zip(beta[0],beta[1])):

        t,b = tup
        if type(t) == str or type(b) == str: continue
        if t > index:
            beta[0][j] +=1
        if b > index:
            beta[1][j] +=1
    beta[0].remove('____')
    beta[1].remove('____')


    for j in range(len(path)):
        if path[j] > index:
            path[j] += 1
path.sort()
for i in range(len(path)):
    index = path[i]
    beta[0][index+1]=int(self.getValue(i,ladder[0],path))
    beta[1][index+1] =int(self.getValue(i,ladder[1],path))
'''
'''
count = {}
count[None] = 0
for i,j in zip(beta[0],beta[1]):
    if i in count.keys():
        count[i] += 1
    else:
        count[i] = 1
    if j in count.keys():
        count[j] += 1
    else:
        count[j] = 1
'''



import numpy as np
from curves import fixMatrixSigns, boundaryCount, genus, ladderConvert, vectorSolution, edges, Three
from graph import Graph

class CurvePair:
    '''
    ladder = None
    beta = None
    top = []
    bottom = []
    n = 0
    matrix = None
    boundaries = None
    genus = None
    edges = None

    solution = None
    distance = 0
    loops = []
    '''
    def __init__(self,topBeta,bottomBeta, dist=1, graph=1):

        is_ladder = lambda top, bottom: not (0 in top or 0 in bottom)

        if is_ladder(topBeta,bottomBeta):
            self.ladder = [topBeta, bottomBeta]
        else:
            self.ladder = None
        


        if is_ladder(topBeta, bottomBeta):
            self.beta = ladderConvert(topBeta, bottomBeta)
            self.top = self.beta[0]
            self.bottom = self.beta[1]
        else:
            self.top = topBeta
            self.bottom = bottomBeta
            self.beta = [self.top, self.bottom]

        self.n = len(self.top)

        self.matrix = np.zeros((2,self.n,4))
        self.matrix[0,:,0] = [self.n-1] + range(self.n-1)
        self.matrix[0,:,1] = self.top
        self.matrix[0,:,2] = range(1,self.n) +[0]
        self.matrix[0,:,3] = self.bottom

        self.matrix = fixMatrixSigns(self.matrix)

        self.boundaries = boundaryCount(self.matrix)
        self.genus = genus(self.matrix)
        self.edges = edges(self.matrix)

        self.solution = vectorSolution(self.edges[0])


        if graph is 1:
            self.loops = Graph(self, self.edges).gammas
        else:
            self.loops = []

        if dist is 1:
            self.distance, self.loopMatrices = self.compute_distance(self.matrix, self.loops)
        else:
            self.distance = None

    def __repr__(self):
        return self.ladder[0]+'\n'+self.ladder[1]+'\n'

    def compute_distance(self, M, allPaths):
        '''
        :param M: the matrix
        :type M:
        :param allPaths:
        :type allPaths:
        :returns: dist: the distance if three/four, or 'Higher' if dist is > 4.

        Computes the distance between the two curves embedded in the matrix.
        If this distance is three, tries to use simple paths to extend the distance
        in a different direction. If this fails, simply returns three;
        else it prints a curve that is distance four from alpha.

        '''

        distIsThree, Lib = Three(M, allPaths)
        #from sys import stderr
        #stderr.write(str(Lib)+'\n')

        dist = 3 if distIsThree  else 'at least 4!'
        return dist, Lib
        '''
        if dist == 3:
            return dist, Lib
        else:
            geodesic_distances = []
            return dist, Lib
            for k,matrix in Lib.iteritems():
                #stderr.write(str(matrix))
                cc_distance = CurvePair(matrix[0, :, 1], matrix[0, :, 3])
                stderr.write(str(k)+": "+str(cc_distance.distance)+'\n')
                geodesic_distances.append(cc_distance.distance)
            return set(geodesic_distances).pop() + 1
        '''



# Paul Glenn
# curves.py
# Give curve pairs class structure in preparation for public access
import numpy as np
from itertools import product
from copy import deepcopy


def fixMatrixSigns(M):
    '''
    :param M: The matrix with incorrect signs.
    :type M: numpy.array(dtype=float64) of shape (2,n,4)
    :returns: The matrix of shape (2,n,4) with appropriate signs.

    Given a matrix corresponding to a curve pair,
    fixes the signs so that the matrix can be traversed by function bdycount.
    Starts from the simples point, then assigns based on matching values and
    ENS: No qualitative results about the curve pair will be changed by this,
    since orientation is just a formality.

    '''

    row, col = [-1,1]
    sgn = 1
    #for ii in range(len(M[0,:])):
    while 0 in M[1,:,1:4:2] and row < M.shape[1]-1:
        row += 1
        sgn = -sgn
        # Starting position and sign are arbitrary.
        for i in range(M.shape[1]):
            gval = int(M[0,row,col])
            new = np.array( np.where(M[0,gval,:]==row) )
            #print  np.where(new%2==col%2)
            ind = int( new[ np.where(new%2==col%2)] )
            sgn = -sgn
            M[1,gval,ind] = sgn
            ind += 2
            ind = ind % M.shape[2]
            sgn = -sgn
            M[1,gval,ind] = sgn
            row = gval
            col = ind
        M[1,:,0] = -1
        M[1,:,2] = 1
    return M


def concatenateMatrix(M1, M2):
    '''
    :param M1: First matrix to combine
    :type M1: numpy.array(dtype=float64) of shape (2,n,4)
    :param M2: Second matrix to combine
    :type M2: numpy.array(dtype=float64) of shape (2,m,4)
    :returns: final matrix

    Note that n does not have to equal m in general.
    Combines two curve pairs into one.
    REQ: curve pairs are, of course, encoded as matrices.
    INV: new matrix signs will be correct
    INV: final matrix will correspond to a single curve.

    '''
    topM = np.copy(M1)
    botM = np.copy(M2)
    n = M1.shape[1]
    m = M2.shape[1]
    topM[0,0,0] = n + m - 1
    topM[0,n-1,2] = n
    botM[0] += n * np.ones(M2[0].shape)
    botM[0,m-1,2] = 0
    botM[0,0,0] = n-1
    newM = np.vstack((topM[0], botM[0])) #  Multi-curve
    newM = np.vstack((newM, np.vstack((topM[1], botM[1])) )).reshape((2,n+m,4))

    left = zip(np.where(topM[0] == n-1)[0], np.where(topM[0] == n-1)[1])
    i0, j0 = 0,0
    for pair in left:
        i, j = pair
        if j%2 == 1: i0,j0 = i,j
    newM[0,i0,j0] = n

    tempInd = np.where(newM[0,n-1,:] == i0)[0]
    tempInd = int(tempInd[tempInd%2==1][0])
    tempVal = newM[0, n-1, tempInd]
    switchVal = newM[0, n, tempInd]
    newM[0, n-1, tempInd] = switchVal
    newM[0,n,tempInd] = tempVal
    returnInd = newM[0,switchVal,:] == n
    returnInd[0:3:2] = False
    newM[0,switchVal,returnInd==True] = n-1

    return fixMatrixSigns(newM)

## Path finding methods -- used to find edge paths in the complement of beta-curve.
## Most are helpers for findAllPaths below.
def visited(curr, face , path):
    '''
    :param curr:
    :type curr:
    :param face:
    :type face:
    :param path:
    :type path:


    '''

    myface = list(face)
    myface.remove(curr)
    v = 0
    for edge in myface:
        if edge in path: v = 1
    return v

def isUnique(path, AllPaths):
    '''
    :param path:
    :type path:
    :param AllPaths:
    :type AllPaths:
    :returns: boolean

    '''
    return not ( path in AllPaths)

def shift(path):
    '''
    Reorders path with a minimum intersection as the base point.
    This is useful for determining path uniqueness.

    :param path: list of intersections representing the current path
    :type path: list
    :returns: list with intersections shifted in order such that the lowest intersection is in the first position.


    '''
    temp = path.index(min(path))
    return path[temp:] + path[:temp]

def invert(path):
    '''
    :param path: some path
    :type path: list
    :returns: the inverted path

    '''
    return shift(path[::-1])

def pathFinishedSingle(edge, face, path):
    '''
    :param edge:
    :type edge:
    :param face:
    :type face:
    :param path:
    :type path:
    :returns:

    '''
    C  = len(path)>2 and (edge == path[-1])
    faceL = list(face)
    faceL.remove(edge)
    C = C and not set(faceL).isdisjoint(set(path))
    #meeting = lambda e1, e2, face: e1 in face and e2 in face
    return C

def pathFinishedDouble(edge, face, path):
    '''
    :param edge:
    :type edge:
    :param face:
    :type face:
    :param path:
    :type path:
    :returns:

    '''

    C  = len(path)>2 and (edge == path[-1])
    faceL = list(face)
    faceL.remove(edge)
    C = C and not set(faceL).isdisjoint(set(path))
    if path.count(edge) != 2 : C = 0
    #meeting = lambda e1, e2, face: e1 in face and e2 in face
    return C

def findNewPaths(currentPath, myface, faces, AllPaths, pathFunction):
    '''
    :param currentPath:
    :type currentPath:
    :param myface:
    :type myface:
    :param faces:
    :type faces:
    :param AllPaths:
    :type AllPaths:
    :param pathFunction:
    :type pathFunction:
    :returns:

    '''



    start = currentPath[0]
    nextEdge = None
    subPath = []
    facesWithoutCurrentFace = list(faces)
    facesWithoutCurrentFace.remove(myface)
    for face in faces: #Check all faces...
        if start in face and face is not myface:
            #if we find the start edge in another face...
            for otherface in facesWithoutCurrentFace:
                #go through its edges...
                if start in otherface: # needed?
                    faceWithoutStart = list(otherface)
                    faceWithoutStart.remove(start)

                    for nonStartingEdge in faceWithoutStart:
                        # try all edges in that path
                        nextEdge = nonStartingEdge
                        if  (not visited( start, otherface, currentPath ) ) and\
                            (nextEdge not in myface):
                            subPath = [nextEdge]
                            subPath.extend(currentPath)

                            # Recursive call to take all possible directions
                            findNewPaths(subPath, otherface, faces, AllPaths, pathFunction)

                        elif pathFunction(nextEdge, otherface, currentPath):
                            newFoundPath = shift(currentPath)
                            invertedPath = invert(currentPath)
                            unique = lambda path: isUnique(path, AllPaths)
                            if unique(newFoundPath) and unique(invertedPath):
                                AllPaths.append(newFoundPath)
                    faceWithoutStart.append(start)
        facesWithoutCurrentFace.append(face)

def removeDuplicates(faces, AllPaths):
    '''
    :param faces:
    :type faces:
    :param AllPaths:
    :type AllPaths:
    :returns:

    '''
    paths = list(AllPaths)
    for path in paths:
        for f in faces:
            counter = 0
            for e in f:
                if e in path: counter += 1
            if counter == 3: AllPaths.remove(path)
    return AllPaths

def findAllPaths(faces):
    '''
    :param faces:
    :type faces:
    :returns:

    '''

    AllPaths = []
    forward = lambda path, face, pathFunction: findNewPaths(path, face, faces, AllPaths, pathFunction)
    for face in faces:
        for edge in face:
            forward([edge], face, pathFinishedSingle)
            #forward([edge], face, pathFinishedDouble)
    AllPaths = removeDuplicates(faces,AllPaths)
    return AllPaths

## Now that we have the paths, they need to be
## re- indexed so that matrices can be built from them.
def buildMatrices(edgePaths, AllPaths):
    '''
    :param edgePaths: list of face boundary orientations
    :type edgePaths: list
    :param AllPaths: list
    :type AllPaths: list

    Take the paths in the skeleton of the complement of the transverse curve
    And create matrices.
    '''
    #print 'edgePaths:',edgePaths
    #print 'AllPaths:',AllPaths[0].loops
    MasterList = []

    # Allow paths to be referenced by face
    #print edgePaths
    for itr in range(len(edgePaths)):
        edgePaths[itr] = dict(edgePaths[itr])

    orderedPaths , mappedPaths = [],[]

    # Rescale path 0-len(path) for matrix
    for Path in AllPaths:
        orderedPath = list(np.sort(Path))
        orderedPaths.append(orderedPath)
        mappedPaths.append(dict(zip(orderedPath,range(len(Path)))))

    #Create Matrices using details from edge paths.
    for Path,mappedPath in zip(AllPaths,mappedPaths):
        #Value Matrix
        pathSize = len(Path) ; shape =(pathSize,4)
        last = Path[len(Path)-1]
        M = np.zeros(shape);
        M2 = np.zeros(shape)
        M[:,0] = np.array([pathSize-1]+range(pathSize-1))
        M[:,2] = np.array(range(1,pathSize)+[0])
        pastEdges = dict(); futureEdges =  dict()
        old_vertex = last

        itr = 1
        for vertex in Path:
            flag = False
            for path in edgePaths:
                keys = path.keys()
                if vertex in keys and old_vertex in keys:
                    pastEdges[vertex] = path[vertex]
                    flag = True
                if vertex in keys and Path[itr%pathSize] in keys:
                    if flag:
                        futureEdges[vertex] = (path[vertex]+ 2) %4
                    else:
                        futureEdges[vertex] = path[vertex]
                flag =False
            old_vertex = vertex
            itr += 1
        old_vertex = last;

        itr = 1
        for vertex in Path:
            curr_vertex = vertex
            next_vertex = Path[itr%pathSize]
            M[mappedPath[vertex],pastEdges[vertex]] = mappedPath[old_vertex]
            M[mappedPath[vertex],futureEdges[vertex]] = mappedPath[next_vertex]
            old_vertex = curr_vertex
            itr += 1

        # Sign matrix: Stand-alone function
        M = fixMatrixSigns(np.array([M, M2], dtype=int))
        MasterList.append(M)
    # End while

    return MasterList

def faceParse(alphaEdges):
    '''
    :param alphaEdges: set of faces with alpha edges.
    :type alphaEdges:
    :returns: (Bridges, Isalnds, lengthCheck) Bridges : 4-sided regions; Islands : n > 4 - gons; lengthCheck : the number of alpha edges included.

    Separate set of all faces into bridges (4-sided regions) and
    islands (higher-sided regions) for distance calculator.

    '''
    Bridges, Islands = [],[]
    lengthCheck = []
    for pair in alphaEdges:
        if pair[0] == 4: Bridges.append(pair)
        else: Islands.append(pair)
        lengthCheck.extend(list(pair[1]))
    return Bridges, Islands, len(lengthCheck)

######					For Distance Extension 					#######
def connected(P1, P2):
    '''
    :param P1:
    :type P1:
    :param P2:
    :type P2:
    :returns:

    '''

    S1  = set(P1)
    S2 = set(P2)
    if S1.isdisjoint(S2): return 0
    elif S1.issubset(S2) and S1.issuperset(S2): return 0
    return 1

def shareEdge(path1, path2):
    '''
    :param path1:
    :type path1:
    :param path2:
    :type path2:
    :returns:

    '''
    share = 0
    if not set(path1).isdisjoint(path2): share = 1
    if share == 0: return 0, -1
    else:
        intersectionSet = set(path1) & set(path2)
        numshared = len(intersectionSet)
        if numshared != 1: return 0, -1 #Then they share too much!
        else:
            sharedItem = intersectionSet.pop()
    return 1,path2.index(sharedItem)

def findCombinedPaths( allPaths, MLibrary):
    '''
    :param allPaths:
    :type allPaths:
    :param MLibrary:
    :type MLibrary:
    :returns:
    '''
    ListofConnected = []
    pathLibrary = dict(zip(range(len(allPaths)),allPaths))
    index1 = 0
    for path1 in allPaths:
        index2 = 0
        for path2 in allPaths:
            if shareEdge(path1, path2)[0]:
                ListofConnected.append((MLibrary[index1],
                            MLibrary[index2]))
            index2 += 1
        index1 += 1

    return ListofConnected

def facesShareTwoEdges(faces):
    '''
    :param faces:
    :type faces:
    :returns:

    '''

    distanceThreeFlag = 0

    for face in faces:
        facesWithoutFace = list(faces)
        facesWithoutFace.remove(face)
        for other in facesWithoutFace:
            if len(set(face) & set(other)) >1: distanceThreeFlag = 1

    return distanceThreeFlag

def edges(M):
    '''
    :param M: the matrix representing a pair of curves
    :type M: numpy.array(dtype=float64) of shape (2,n, 4)
    :returns: (allFaces, edges) allFaces: tuple of faces including size and set of alpha-edges which bound them; edges: same as allFaces, exceincluding orientation of boundary edges.
    '''
    #print 'called edges'
    # The em list is needed to hold the tuples of (faceLength, faceEdges)
    allFaces = [ ]
    # INV: Number of faces found.
    faces = 0 ;

    numrows, numcols = M.shape[1:3] #num_rows, num_cols

    old_Vertices = [ ] ##list of previous paths
    bigonFlag = 0
    # Bigons are unwanted structures.
    Paths = [ ]  # alpha - edge paths.
    facesTemp = dict()
    for i,j in product(range(numrows),range(numcols)):
        #Set of edges associated with face
        tr = 1
        face = set()
        if faces==numrows: break #upper bound on possible no. faces
        # Start position in matrix: returning to this means a face has been
        # enclosed
        io =i
        jo=j

        found = 0 #exit condition
        # Number of edges for face. Keeps track of vector solution
        edges=0
        pathTemp = [];

        # Begin traversal
        while not found:

            gval = int(M[0,i,j]);
            #current value at index gives next vertex/row
            #value check
            arr1 = np.where(M[0,gval,:] == i%numrows)
            arr2 = np.where(M[1,gval,:] != M[1,i,j]) #sign check

            i_next = gval
            alpha = (M[0,i,0]+1)%numrows
            i= i_next

            new = np.intersect1d(arr1[0],arr2[0],assume_unique=True)
            #ENS: val and sign correct
            ind = np.where(new%2==j%2) #ENS: beta->beta, alpha->alpha
            j_next = (int(new[ind])+1)%numcols #Always move clockwise
            j_old = j
            j = j_next
            if (i,j) in old_Vertices:
                break
            old_Vertices.append((i,j))
            edges += 1

            alpha_new =( M[0,i,0]+1) % numrows;
            shift = (alpha_new - alpha)

            if shift==1 and j%2 == 1:
                face.add(alpha_new)
                pathTemp.append((alpha_new,(j)%numcols))
            elif shift==-1 and j%2==1:
                face.add(alpha)
                pathTemp.append((alpha,(j)%numcols))
            elif shift == numrows-1 and alpha ==0  and j%2==1: #
                face.add(alpha)
                pathTemp.append((alpha,(j)%numcols))
            elif shift == 1-numrows and alpha_new==0 and j%2==1: #
                face.add(alpha_new)
                pathTemp.append((alpha_new,(j)%numcols))

            if (i,j)==(io,jo):
                facesTemp[edges] = facesTemp.get(edges,0) +1
                if edges==2:
                    bigonFlag = 1
                if not bigonFlag:
                    Paths.append(pathTemp)
                    faces += 1
                    found = 1
                    allFaces.append((edges,face))
    return allFaces, Paths

def boundaryCount(M):
    '''
    :param M: the matrix representing a pair of curves
    :type M: numpy.array(dtype=float64) of shape (2,n, 4)
    :returns: (faces, bigon) faces: number of faces; bigon: 1 iff a bigon is found

    'Simply' count the number of faces bounded by two filling curves on a surface.
    Curves must be encoded as matrix according to vertices of intersection and
    associated orientation.


    '''

    faces = 0
    numrows, numcols = M.shape[1:3] #num_rows, num_cols
    oldEdges = [ ] ##list of previous edge paths
    bigonFlag =0

    for i,j in product(range(numrows),range(numcols)):
        # upper bound on possible no. faces is number of vertices
        if faces==numrows: break

        io =i
        jo=j
        #First matrix element; will go to vertex M[0,i,j]

        found = 0 #Exit condition
        pathLength=0 # Keep track of path length

        while not found:
            gval = int(M[0,i,j]);
            #current value at index gives next vertex/row

            arr1 = np.where(M[0,gval,:] == i%numrows) #value check
            arr2 = np.where(M[1,gval,:] == -M[1,i,j]); #sign should flip +/-
            i = gval # Go to next vertex/row
            new = np.intersect1d(arr1[0],arr2[0],assume_unique=True)
            #ENS: val and sign correct

            ind = np.where(new%2==j%2)
            #ENS: beta->beta, alpha->alpha

            j = (int(new[ind])+1)%numcols
            #Always move clockwise - to next edge

            if (i,j) in oldEdges: break
            oldEdges.append((i,j))
            # To save work and not go on old paths.
            # Also so we don't count faces twice...
            pathLength += 1
            # The path length is the number of edges traversed in the current face.

            if (i,j)==(io,jo):
                if pathLength==2: bigonFlag = 1;
                # Two edges to a face --> bigon
                faces += 1
                found = 1 #INV: found = 1 -> has found a bdy curve
    return faces, bigonFlag

def vectorSolution(edges):

    solution = dict()

    for face in edges:
        if face[0] not in solution.keys():
            solution[face[0]] = 1
        else:
            solution[face[0]] += 1

    return solution

def genus(M, euler=0, boundaries = 0):
    '''
    :param M: the matrix representing a pair of curves
    :type M: numpy.array(dtype=float64) of shape (2,n, 4)
    :param euler: 0 if euler characteristic not needed, else 1
    :type euler: int

    :returns: (g,X): g: genus; X : euler characteristic

    Compute the genus of the span of a curve pair, i.e. the minimal genus surface
    on which they fill.

    '''
    V = M.shape[1] # vertices
    P, bigon = boundaryCount(M) # Polygons in the complement of curve pair
    #if bigon is 1: P -= 1 # pull away one bigon
    #Euler characteristic (since edges = 2*vertices) is P - V
    # originally X = V-E+P
    X = P-V+boundaries

    # genus = 1 - 0.5*euler_characteristic
    Genus =  (2-X)/2
    returnVal = dict([(0,Genus),(1,(Genus,X))])
    # For return purposes only.

    if bigon: Genus -= 1 # Bigons steal genus; this gives it back.

    return returnVal[euler]


def testCollection(matrixList, originalGenus):
    '''
    :param matrixList:
    :type matrixList: list of numpy.array(dtype=float64)
    :param originalGenus:
    :type originalGenus:
    :returns: list of booleans and a dictionary containing the matrices without bigons.

    Test a collection of matrices to see if they fill a given genus.
    Note: Matrices with bigons are still counted in the calculation, but are not returned.
    '''
    genusCollection = []
    index = 0
    matrixLibrary = dict()
    for M in matrixList:

        bigon = boundaryCount(M)[1]
        Genus = genus(M)
        matrixLibrary[index] = M
        index += 1
        if bigon:
            Genus  = 0 # If it has a bigon, it should automatically fail

        genusCollection.append(Genus)
    # Test if all fill and thus distance g.t. 3
    test = [Genus == originalGenus for Genus in genusCollection]
    return test, matrixLibrary

def fourgonTest(F4, Fn):
    '''
    :param F4:
    :type F4:
    :param Fn:
    :type Fn:
    :returns: boolean: True if Fn is a 4-gon, False otherwise

    '''

    for islandFace in Fn:
        for bridgeFace in F4:
            if len(islandFace[1] & bridgeFace[1]) >= 2: return True
    return False

def Three(M, allPaths, ext=0):
    '''
    :param M:
    :type M:
    :param allPaths:
    :type allPaths:
    :param ext:
    :type exr:
    :returns: 1 if the curve pair is distance three and 0 otherwise.

    '''

    three, matrixLibrary = 0, dict()


    F0, bigon = boundaryCount(M)
    originalGenus = genus(M)
    #Calculate face alpha edges and alpha edge paths
    Faces, edgePaths = edges(M)
    Bridges, Islands, lengthCheck = faceParse(Faces)

    # Bridges are faces bounded by four edges.
    # Islands are bounded bt more than four, say six or eight...
    bridgeFaces = [list(face[1]) for face in Bridges]
    islandFaces = [list(face[1]) for face in Islands]

    ############# Quick and dirty checks for distance three #############
    if F0 == 1 :
        three = 1
    #	print '''The complement contains a polygon which shares an edge with
    #			itself, and so is distance 3. '''


    # If any face is larger than the number of vertices, it will definitely
    # Share an edge with itself. Therefore distance three.
    faceSizeTest = [face[0]  > M.shape[1] for face in Faces]
    if True in faceSizeTest:
        three = 1
    #	print '''The complement contains a polygon which shares an edge with
    #			itself, and so is distance 3. '''


    if  fourgonTest(Bridges, Islands):
        three = 1


    if lengthCheck != 2*M.shape[1]:
        three = 1
    #	print '''The complement contains a polygon which shares an edge with
    #			itself, and so is distance 3. '''
    # See function faceParse
    # If this is true, then some face is sharing an edge with itself.
    # It can't be a four-gon, since that would mean we have a multi-curve.
    # So it must be an island, which means that there is a curve
    # in the complement which is distance two. Thus the original pair is
    # Distance three.

    if facesShareTwoEdges(islandFaces):
        three = 1
    #	print ''' Found two faces that share multiple edges: A curve that
    #			  intersects the non-reference curve only two times has been
    #			  found. This curve cannot fill and so the pair is distance 3.
    #		  '''

    # Means there is a path of length two.
    # Since a curve pair with two intersections cannot fill on any genus >1 ,
    # The two curves will be distance three.

    #Find linking edge paths in "mesh"
    faces  = list(bridgeFaces)
    faces += islandFaces

    #Build paths into curves and intersect with alpha
    matrixList = buildMatrices(edgePaths, allPaths)

    genusTest, matrixLibrary = testCollection(matrixList,originalGenus)
    if three != 1:
        three = 1 if False in genusTest else 0

    # Output
    returnVals = [three, matrixLibrary]
    return returnVals

def ladderConvert(ladderTop, ladderBottom):
    #print ladderTop, ladderBottom
    n = len(ladderTop)
    newTop = [' ']*n
    newBottom = list(newTop)

    for j in range(1, n+1):

        if j in ladderTop and j in ladderBottom:
            newTop[ladderTop.index(j)]  = ladderBottom.index(j)
            newBottom[ladderBottom.index(j)] = ladderTop.index(j)

        elif j in ladderTop:
            ladderTopTemp = list(ladderTop)
            indices = [ladderTop.index(j)]; ladderTopTemp[indices[0]] = None#ladderTopTemp.remove(j)
            indices.append(ladderTopTemp.index(j))
            newTop[indices[0]] = indices[1]
            newTop[indices[1]] = indices[0]

        elif j in ladderBottom:
            ladderBottomTemp = list(ladderBottom)
            indices = [ladderBottom.index(j)]; ladderBottomTemp[indices[0]] = None #ladderBottomTemp.remove(j)
            indices.append(ladderBottomTemp.index(j))
            newBottom[indices[0]] = indices[1]
            newBottom[indices[1]] = indices[0]

    return newTop, newBottom


def ladder_is_multicurve(top, bottom):

    n = len(top)

    j0 = top[0]
    counter = 1

    j = bottom[0]
    bottom[0] = None
    oldIndex = 0
    while j != j0:

        old_j = j
        if j in top:
            nextIndex = top.index(j)
            j = bottom[nextIndex]

            if None in top:
                top[top.index(None)] =  old_j
            elif None in bottom:
                bottom[bottom.index(None)] =  old_j

            bottom[nextIndex] = None

        elif j in bottom:
            nextIndex = bottom.index(j)
            j = top[nextIndex]

            if None in top:
                top[top.index(None)] =  old_j
            elif None in bottom:
                bottom[bottom.index(None)] =  old_j

            top[nextIndex] = None

        counter += 1

    if None in top:
        top[top.index(None)] =  j
    elif None in bottom:
        bottom[bottom.index(None)] =  j


    return 1 if counter != n else 0

def matrix_is_multicurve(beta):

    top = beta[0]
    bottom = beta[1]
    index = 0
    j = top[index]
    start = 100
    counter = 0

    while j != start:
        counter += 1
        start = top[0]

        if top[j] == index :
            next_index = j
            j = bottom[j]
            index = next_index

        elif bottom[j] == index:
            next_index = j
            j = top[j]
            index = next_index

        if top[j] == bottom[j]: return True
        #print 'j', j,'next_index: ',  index

    return False if counter == len(top) else True


def test_permutations(original_ladder):
    distance4 = []
    distance3 = []
    from curvepair import CurvePair
    ladder = deepcopy(original_ladder)
    if not original_ladder is None:
        for i in range(len(ladder[0])):
            if not ladder_is_multicurve(*ladder):
                perm = CurvePair(*ladder)
            else:
                perm = None
            if not perm is None :
                if perm.distance is 4:
                    distance4.append(deepcopy(perm))
                else:
                    distance3.append(deepcopy(perm))
            else: pass
            first_vertex = ladder[0].pop(0)
            ladder[0].append(first_vertex)

        if len(distance4) == 0:
            print ' Found no distance four permutations of the ladder. '

        if distance3:
            print 'Distance 3 single curves: '
            for curve in distance3:
                print 'top   : ', curve.ladder[0]
                print 'bottom: ', curve.ladder[1]
        if distance4:
            print 'Distance 4+ single curves: '
            for curve in distance4:
                print 'top   : ', curve.ladder[0]
                print 'bottom: ', curve.ladder[1]

        return distance4
    else:
        print "You didn't give me a ladder! "
        return []


def test_perms(original_ladder):
    distance4 = []
    distance3 = []
    ladder_to_perm = deepcopy(original_ladder)
    from curvepair import CurvePair
    if not original_ladder is None:
        for i in range(len(ladder_to_perm[0])):
            if not ladder_is_multicurve(*ladder_to_perm):
                perm = CurvePair(*deepcopy(ladder_to_perm))
            else:
                perm = None
            if not perm is None :
                if not perm.distance is 3:
                    distance4.append(perm)
                else:
                    distance3.append(perm)
            else: pass

            first_vertex = ladder_to_perm[0].pop(0)
            ladder_to_perm[0].append(first_vertex)
    return distance3, distance4
